#include "App.h"
#include "VoltageMonitor.h"
#include "TemperatureSensor.h"
#include "Radio.h"

#include "stm32l0xx_hal.h"
#include "stm32l0xx_ll_utils.h"
#include "main.h"
#include "wolfSSL.I-CUBE-wolfSSL_conf.h"
#include <wolfssl/wolfcrypt/aes.h>

#include <string.h>

extern RTC_HandleTypeDef hrtc;
extern RNG_HandleTypeDef hrng;
void SystemClock_Config(void);

static const uint8_t crc8Table[] = {
		0x00, 0x07, 0x0e, 0x09, 0x1c, 0x1b, 0x12, 0x15,
		0x38, 0x3f, 0x36, 0x31, 0x24, 0x23, 0x2a, 0x2d,
		0x70, 0x77, 0x7e, 0x79, 0x6c, 0x6b, 0x62, 0x65,
		0x48, 0x4f, 0x46, 0x41, 0x54, 0x53, 0x5a, 0x5d,
		0xe0, 0xe7, 0xee, 0xe9, 0xfc, 0xfb, 0xf2, 0xf5,
		0xd8, 0xdf, 0xd6, 0xd1, 0xc4, 0xc3, 0xca, 0xcd,
		0x90, 0x97, 0x9e, 0x99, 0x8c, 0x8b, 0x82, 0x85,
		0xa8, 0xaf, 0xa6, 0xa1, 0xb4, 0xb3, 0xba, 0xbd,
		0xc7, 0xc0, 0xc9, 0xce, 0xdb, 0xdc, 0xd5, 0xd2,
		0xff, 0xf8, 0xf1, 0xf6, 0xe3, 0xe4, 0xed, 0xea,
		0xb7, 0xb0, 0xb9, 0xbe, 0xab, 0xac, 0xa5, 0xa2,
		0x8f, 0x88, 0x81, 0x86, 0x93, 0x94, 0x9d, 0x9a,
		0x27, 0x20, 0x29, 0x2e, 0x3b, 0x3c, 0x35, 0x32,
		0x1f, 0x18, 0x11, 0x16, 0x03, 0x04, 0x0d, 0x0a,
		0x57, 0x50, 0x59, 0x5e, 0x4b, 0x4c, 0x45, 0x42,
		0x6f, 0x68, 0x61, 0x66, 0x73, 0x74, 0x7d, 0x7a,
		0x89, 0x8e, 0x87, 0x80, 0x95, 0x92, 0x9b, 0x9c,
		0xb1, 0xb6, 0xbf, 0xb8, 0xad, 0xaa, 0xa3, 0xa4,
		0xf9, 0xfe, 0xf7, 0xf0, 0xe5, 0xe2, 0xeb, 0xec,
		0xc1, 0xc6, 0xcf, 0xc8, 0xdd, 0xda, 0xd3, 0xd4,
		0x69, 0x6e, 0x67, 0x60, 0x75, 0x72, 0x7b, 0x7c,
		0x51, 0x56, 0x5f, 0x58, 0x4d, 0x4a, 0x43, 0x44,
		0x19, 0x1e, 0x17, 0x10, 0x05, 0x02, 0x0b, 0x0c,
		0x21, 0x26, 0x2f, 0x28, 0x3d, 0x3a, 0x33, 0x34,
		0x4e, 0x49, 0x40, 0x47, 0x52, 0x55, 0x5c, 0x5b,
		0x76, 0x71, 0x78, 0x7f, 0x6a, 0x6d, 0x64, 0x63,
		0x3e, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2c, 0x2b,
		0x06, 0x01, 0x08, 0x0f, 0x1a, 0x1d, 0x14, 0x13,
		0xae, 0xa9, 0xa0, 0xa7, 0xb2, 0xb5, 0xbc, 0xbb,
		0x96, 0x91, 0x98, 0x9f, 0x8a, 0x8d, 0x84, 0x83,
		0xde, 0xd9, 0xd0, 0xd7, 0xc2, 0xc5, 0xcc, 0xcb,
		0xe6, 0xe1, 0xe8, 0xef, 0xfa, 0xfd, 0xf4, 0xf3
};

uint8_t iv[16] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
uint8_t ivKey[16] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
uint8_t aesKey[16] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

void App_SleepLong() {
	HAL_PWREx_EnableUltraLowPower();
	HAL_PWR_EnterSTOPMode(PWR_LOWPOWERREGULATOR_ON, PWR_STOPENTRY_WFI);

	// wake up
	SystemClock_Config();
}

void App_SleepShortTicks(int ticks) {
	HAL_ResumeTick();
	HAL_Delay(ticks);
	HAL_SuspendTick();
}

void App_SleepShortMs(int ms) {
	HAL_ResumeTick();
	HAL_Delay(ms);
	HAL_SuspendTick();
}

static uint16_t App_GenerateBootId() {

	__HAL_RCC_RNG_CLK_ENABLE();
	MX_RNG_Init();

	uint32_t randomNumber;

	HAL_StatusTypeDef hStatus = -1;
	uint32_t attemps = 100;
	while (hStatus != 0 && attemps--) {
		hStatus = HAL_RNG_GenerateRandomNumber(&hrng, &randomNumber);
		if (hStatus) {
			App_SleepShortMs(1);
		}
	}

	HAL_RNG_DeInit(&hrng);
	__HAL_RCC_RNG_FORCE_RESET();
	__HAL_RCC_RNG_RELEASE_RESET();
	__HAL_RCC_RNG_CLK_DISABLE();

	if (hStatus)  {
		return 0x7977;
	} else {
		return (uint16_t)randomNumber;
	}

}

static void App_EncodeMessage(uint8_t* message, int32_t temperatureSum, int temperaturesCount, int16_t temperatureMin, int16_t temperatureMax, uint16_t voltage10, uint16_t voltage20, uint16_t voltage30, uint32_t deviceId, uint16_t bootId, uint16_t packetNo) {
	temperatureSum /= temperaturesCount;
	if (temperatureSum < INT16_MIN) {
		temperatureSum = INT16_MIN;
	} else if (temperatureSum > INT16_MAX) {
		temperatureSum = INT16_MAX;
	}

	int16_t temperatureAvg = temperatureSum;

	uint32_t headerVoltage1 =
			(((uint32_t)voltage30) << 0) |
			(((uint32_t)voltage20) << 12) |
			(((uint32_t)voltage10) << 24);

	uint16_t headerVoltage2 = (voltage10 >> 8);

	memcpy(message + 0, &deviceId, 4);
	memcpy(message + 4, &bootId, 2);
	memcpy(message + 6, &packetNo, 2);
	memcpy(message + 8, &headerVoltage2, 2);
	memcpy(message + 10, &temperatureAvg, 2);
	memcpy(message + 12, &temperatureMin, 2);
	memcpy(message + 14, &temperatureMax, 2);
	memcpy(message + 16, &headerVoltage1, 4);

	uint8_t crc = 0x00;
	for (int i = 0; i < 20; i++) {
		crc = crc8Table[crc ^ message[i]];
	}

	headerVoltage2 |= ((uint16_t)crc) << 8;
	memcpy(message + 8, &headerVoltage2, 2);

	Aes aes;
	int aesStatus;
	aesStatus = wc_AesInit(&aes, NULL, INVALID_DEVID);
	if (aesStatus) {
		wc_AesFree(&aes);
		__BKPT(0);
	}

	aesStatus = wc_AesSetIV(&aes, iv);
	if (aesStatus) {
		wc_AesFree(&aes);
		__BKPT(0);
	}

	aesStatus = wc_AesSetKey(&aes, aesKey, sizeof(aesKey), ivKey, AES_ENCRYPTION);
	if (aesStatus) {
		wc_AesFree(&aes);
		__BKPT(0);
	}

	uint8_t encrypted[16];

	aesStatus = wc_AesCbcEncrypt(&aes, encrypted, message + 4, 16);
	if (aesStatus) {
		wc_AesFree(&aes);
		__BKPT(0);
	}

	wc_AesFree(&aes);

	memcpy(message + 4, encrypted, 16);
}

void App_Run() {
	int temperatureStatus;
	int voltageStatus;
	int transmitStatus;

	// HAL_SuspendTick()
	// HAL_SuspendTick() is called inside App_SleepShortMs(2);

	// wait for chips powerup
	App_SleepShortMs(2);

	App_GenerateBootId();

	Radio_Init();

	__HAL_RCC_GPIOH_CLK_DISABLE();

	uint32_t deviceId = LL_GetUID_Word0() ^ LL_GetUID_Word1() ^ LL_GetUID_Word2();
	uint16_t bootId = App_GenerateBootId();
	uint16_t packetNo = 0;

	while (1) {
		int32_t temperatureSum = 0;
		int temperaturesCount = 0;
		int16_t temperatureMin = INT16_MAX;
		int16_t temperatureMax = INT16_MIN;

		uint16_t voltage30 = 0xFFFF;
		uint16_t voltage20 = 0xFFFF;
		uint16_t voltage10 = 0xFFFF;

		for (int i = 0; i < 30; i++) {
			temperatureStatus = TemperatureSensor_BeginMeasurement();
			if (temperatureStatus) {
				App_SleepLong();
				continue;
			}

			VoltageMonitor_Voltages voltages;
			voltageStatus = VoltageMonitor_GetVoltages(&voltages);

			int16_t temperature;
			temperatureStatus = TemperatureSensor_GetTemperature(&temperature);
			if (temperatureStatus) {
				App_SleepLong();
				continue;
			}

			if (voltageStatus == 0) {
				if (i >= 10 && i <= 15 && voltage10 == 0xFFFF) {
					voltage10 = voltages.batteryVoltage;
				} else if (i >= 20 && i <= 25 && voltage20 == 0xFFFF) {
					voltage20 = voltages.batteryVoltage;
				} else if (i > 25) {
					voltage30 = voltages.batteryVoltage;
				}
			}

			temperatureSum += temperature;
			temperaturesCount++;
			if (temperature < temperatureMin) {
				temperatureMin = temperature;
			}
			if (temperature > temperatureMax) {
				temperatureMax = temperature;
			}

			// in last iteration we do not sleep because we need to transmit data before sleep
			if (i != 29) {
				App_SleepLong();
			}
		}

		uint8_t message[20];
		App_EncodeMessage(message, temperatureSum, temperaturesCount, temperatureMin, temperatureMax, voltage10, voltage20, voltage30, deviceId, bootId, packetNo);

		packetNo++;
		if (packetNo == 0) {
			bootId = App_GenerateBootId();
		}


		transmitStatus = Radio_TransmitMessage(message, sizeof(message));
		if (transmitStatus) {
			__BKPT(0);
		}

		App_SleepLong();
	}

}
