#include "Radio.h"
#include "App.h"

#include "main.h"
#include "stm32l0xx_hal.h"

extern SPI_HandleTypeDef hspi1;
extern TIM_HandleTypeDef htim6;

static void Radio_ReInitGpio() {
	// configure MAX41460 input pins to driven value (reduces MAX41460 power consumption) before first use
	GPIO_InitTypeDef gpio;
	gpio.Alternate = 0;
	gpio.Mode = GPIO_MODE_OUTPUT_PP;
	gpio.Pull = GPIO_NOPULL;
	gpio.Speed = GPIO_SPEED_LOW;

	gpio.Pin = RADIO_MOSI_Pin;
	HAL_GPIO_WritePin(RADIO_MOSI_GPIO_Port, RADIO_MOSI_Pin, 0);
	HAL_GPIO_Init(RADIO_MOSI_GPIO_Port, &gpio);

	gpio.Pin = RADIO_SCLK_Pin;
	HAL_GPIO_WritePin(RADIO_SCLK_GPIO_Port, RADIO_SCLK_Pin, 0);
	HAL_GPIO_Init(RADIO_SCLK_GPIO_Port, &gpio);
}

void Radio_Init() {
	Radio_ReInitGpio();

	// cycle MAX41460 DATA pin according to "Power-On Programming" section in MAX41460 datasheet rev. 2
	HAL_GPIO_WritePin(RADIO_MOSI_GPIO_Port, RADIO_MOSI_Pin, 0);
	App_SleepShortMs(1);
	HAL_GPIO_WritePin(RADIO_MOSI_GPIO_Port, RADIO_MOSI_Pin, 1);
	App_SleepShortMs(1);
	HAL_GPIO_WritePin(RADIO_MOSI_GPIO_Port, RADIO_MOSI_Pin, 0);
	App_SleepShortMs(1);
	HAL_GPIO_WritePin(RADIO_MOSI_GPIO_Port, RADIO_MOSI_Pin, 1);
	App_SleepShortMs(1);
	HAL_GPIO_WritePin(RADIO_MOSI_GPIO_Port, RADIO_MOSI_Pin, 0);
}


static void Radio_PowerOn() {
	// Init SPI Driver
	__HAL_RCC_SPI1_CLK_ENABLE();
	MX_SPI1_Init();

	// init TIM6 Driver
	__HAL_RCC_TIM6_CLK_ENABLE();
	MX_TIM6_Init();
}

static void Radio_PowerOff() {
	HAL_StatusTypeDef hStatus;

	HAL_GPIO_WritePin(RADIO_CSB_GPIO_Port, RADIO_CSB_Pin, 1);

	hStatus = HAL_SPI_DeInit(&hspi1);
	// error ignored
	(void) hStatus;

	// shutdown SPI peripheral
	__HAL_RCC_SPI1_FORCE_RESET();
	__HAL_RCC_SPI1_RELEASE_RESET();
	__HAL_RCC_SPI1_CLK_DISABLE();

	hStatus = HAL_TIM_Base_DeInit(&htim6);
	// error ignored
	(void) hStatus;

	// shutdown SPI peripheral
	__HAL_RCC_TIM6_FORCE_RESET();
	__HAL_RCC_TIM6_RELEASE_RESET();
	__HAL_RCC_TIM6_CLK_DISABLE();

	// reinitialize GPIOs
	Radio_ReInitGpio();
}

#define MAX41460_WriteRegExitCs(addr, val, csbClr, csbSet) \
	RADIO_CSB_GPIO_Port->ODR = csbClr; \
	\
	SPI1->DR = addr; \
	while ((SPI1->SR & SPI_SR_TXE) == 0) {} \
	SPI1->DR = val; \
	while ((SPI1->SR & SPI_SR_TXE) == 0) {} \
	\
	RADIO_CSB_GPIO_Port->ODR = csbSet;

#define MAX41460_WriteReg(addr, val, csbClr) \
	RADIO_CSB_GPIO_Port->ODR = csbClr; \
	\
	SPI1->DR = addr; \
	while ((SPI1->SR & SPI_SR_TXE) == 0) {} \
	SPI1->DR = val; \
	while ((SPI1->SR & SPI_SR_TXE) == 0) {}

#define MAX41460_WriteByte(data) \
	SPI1->DR = data; \
	while ((SPI1->SR & SPI_SR_TXE) == 0) {}

/*
 static void Radio_Delay(uint16_t prescaler, uint16_t targetCount) {
 TIM6->PSC = prescaler - 1;
 TIM6->CNT = 0;
 TIM6->CR1 |= TIM_CR1_CEN;

 while (TIM6->CNT < targetCount) {
 }

 TIM6->CR1 &= ~TIM_CR1_CEN;
 }
 */

#define Radio_Delay15Us __NOP(); __NOP(); __NOP

#define Radio_Delay125Us \
	__NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); \
	__NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); \
	__NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); \
	__NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); \
	__NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); \
	__NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP();  \
	__NOP

static void MAX41460_InitialProgramming() {
	// forward GPIO preparation
	HAL_GPIO_WritePin(RADIO_MOSI_GPIO_Port, RADIO_MOSI_Pin, 0);
	uint32_t mosiModeChangeValue = 0;
	for (int i = 0; i < 16; i++) {
		if ((1 << i) == RADIO_MOSI_Pin) {
			mosiModeChangeValue = (RADIO_MOSI_GPIO_Port->MODER & (~(0x3 << (i * 2)))) | (0x1 << (i * 2));
		}
	}

	SPI1->CR1 |= SPI_CR1_SPE;

	// transmit dummy byte with CS=1 to stabilize clock low
	SPI1->DR = 0;
	while ((SPI1->SR & SPI_SR_TXE) == 0) {}

	uint32_t csbClr = RADIO_CSB_GPIO_Port->ODR & ~(RADIO_CSB_Pin);
	uint32_t csbSet = RADIO_CSB_GPIO_Port->ODR | (RADIO_CSB_Pin);


	// assert MAX41460 reset
	Radio_Delay15Us();
	RADIO_CSB_GPIO_Port->ODR = csbClr;
	Radio_Delay125Us();
	Radio_Delay125Us();
	MAX41460_WriteByte(MAX41460_REG_CFG8);
	MAX41460_WriteByte(0x01); /* MAX41460_REG_CFG8 */
	RADIO_CSB_GPIO_Port->ODR = csbSet;
	Radio_Delay125Us();


	// deassert MAX41460 reset
	RADIO_CSB_GPIO_Port->ODR = csbClr;
	Radio_Delay125Us();
	Radio_Delay125Us();
	MAX41460_WriteByte(MAX41460_REG_CFG8);
	MAX41460_WriteByte(0x00); /* MAX41460_REG_CFG8 */
	RADIO_CSB_GPIO_Port->ODR = csbSet;
	Radio_Delay125Us();

	// initial programming
	RADIO_CSB_GPIO_Port->ODR = csbClr;
	Radio_Delay125Us();
	Radio_Delay125Us();


	MAX41460_WriteByte(MAX41460_REG_CFG1); // write bit + register address
	MAX41460_WriteByte(0x90); /* MAX41460_REG_CFG1 */
	MAX41460_WriteByte(0x81); /* MAX41460_REG_CFG2 */
	MAX41460_WriteByte(0x03); /* MAX41460_REG_CFG3 */
	MAX41460_WriteByte(0x00); /* MAX41460_REG_CFG4 */
	MAX41460_WriteByte(0x00); /* MAX41460_REG_CFG5 */
	MAX41460_WriteByte(0x04); /* MAX41460_REG_SHDN */
	MAX41460_WriteByte(0x85); /* MAX41460_REG_PA1  */
	MAX41460_WriteByte(0x80); /* MAX41460_REG_PA2  */
	MAX41460_WriteByte(0x60); /* MAX41460_REG_PLL1 */
	MAX41460_WriteByte(0x00); /* MAX41460_REG_PLL2 */
	MAX41460_WriteByte(0x00); /* MAX41460_REG_CFG6 */


	// 868.40 MHz (in CZE 1% Duty Cycle Limit)
	MAX41460_WriteByte(0x36); /* MAX41460_REG_PLL3 */
	MAX41460_WriteByte(0x48); /* MAX41460_REG_PLL4 */
	MAX41460_WriteByte(0x66); /* MAX41460_REG_PLL5 */

	// 869.50 MHz (in CZE 10% Duty Cycle Limit)
	// MAX41460_WriteByte(0x36); /* MAX41460_REG_PLL3 */
	// MAX41460_WriteByte(0x58); /* MAX41460_REG_PLL4 */
	// MAX41460_WriteByte(0x00); /* MAX41460_REG_PLL5 */

	MAX41460_WriteByte(0x28); /* MAX41460_REG_PLL6 */
	MAX41460_WriteByte(0x04); /* MAX41460_REG_PLL7 */

	// Tpll
	Radio_Delay125Us();

	RADIO_CSB_GPIO_Port->ODR = csbSet;

	// time since start > Tpll_MICROSECONDS + Txo_MICROSECONDS
	// time since start > 333uS
	Radio_Delay125Us();

	RADIO_CSB_GPIO_Port->ODR = csbClr;
	Radio_Delay125Us();
	Radio_Delay125Us();

	// write bit + register address
	MAX41460_WriteByte(MAX41460_REG_CFG7);
	MAX41460_WriteByte(0x02); /* MAX41460_REG_CFG7 */

	/*
	MAX41460_WriteRegExitCs(MAX41460_REG_CFG1, 0x90, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_CFG2, 0x81, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_CFG3, 0x03, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_CFG4, 0x00, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_CFG5, 0x00, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_SHDN, 0x04, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_PA1, 0x80, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_PA2, 0x80, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_PLL1, 0x60, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_PLL2, 0x00, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_CFG6, 0x00, csbClr, csbSet);

	// 433.92 MHz
	// MAX41460_WriteRegExitCs(MAX41460_REG_PLL3, 0x1B, csbClr, csbSet);
	// MAX41460_WriteRegExitCs(MAX41460_REG_PLL4, 0x1E, csbClr, csbSet);
	// MAX41460_WriteRegExitCs(MAX41460_REG_PLL5, 0xB8, csbClr, csbSet);

	// 868.40 MHz (in CZE 1% Duty Cycle Limit)
	// MAX41460_WriteRegExitCs(MAX41460_REG_PLL3, 0x36, csbClr, csbSet);
	// MAX41460_WriteRegExitCs(MAX41460_REG_PLL4, 0x48, csbClr, csbSet);
	// MAX41460_WriteRegExitCs(MAX41460_REG_PLL5, 0x66, csbClr, csbSet);

	// 869.50 MHz (in CZE 10% Duty Cycle Limit)
	MAX41460_WriteRegExitCs(MAX41460_REG_PLL3, 0x36, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_PLL4, 0x58, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_PLL5, 0x00, csbClr, csbSet);

	MAX41460_WriteRegExitCs(MAX41460_REG_PLL6, 0x28, csbClr, csbSet);
	MAX41460_WriteRegExitCs(MAX41460_REG_PLL7, 0x04, csbClr, csbSet);

	// Tpll_MICROSECONDS
	// 125us
	// Radio_Delay(66, 1);
	Radio_Delay125Us();

	MAX41460_WriteReg(MAX41460_REG_CFG7, 0x02, csbClr);
	*/

	// switch MISO control from SPI to GPIO
	RADIO_MOSI_GPIO_Port->MODER = mosiModeChangeValue;

	// Ttx_MICROSECONDS
	// 15us
	Radio_Delay15Us();
}

int Radio_TransmitMessage(uint8_t *message, size_t len) {
	if (len > RADIO_TRANSMIT_MAX_LEN_BYTES) {
		return 1;
	}

	uint32_t odr = RADIO_MOSI_GPIO_Port->ODR & (~RADIO_MOSI_Pin);
	if (RADIO_MOSI_GPIO_Port == RADIO_CSB_GPIO_Port) {
		odr &= ~RADIO_CSB_Pin;
	}

	// + RADIO_TRANSMIT_TERMINATION_PULSES * 2
	uint32_t odrs[RADIO_TRANSMIT_PREAMBLE_PULSES * 2 + RADIO_TRANSMIT_SYNC_BYTES * 8 + RADIO_TRANSMIT_MAX_LEN_BYTES * 8 * 2];
	uint32_t* odrWr = odrs;
	for (size_t i = 0; i < RADIO_TRANSMIT_PREAMBLE_PULSES; i++) {
		*odrWr++ = odr | RADIO_MOSI_Pin;
		*odrWr++ = odr;
	}
	for (int i = RADIO_TRANSMIT_SYNC_BYTES * 8; i >= 0; i--) {
		if (0x930B51DE & (1 << i)) {
			*odrWr++ = odr | RADIO_MOSI_Pin;
		} else {
			*odrWr++ = odr;
		}
	}
	for (size_t i = 0; i < len; i++) {
		for (int j = 7; j >= 0; j--) {
			// manchester encoding

			if (message[i] & (1 << j)) {
				*odrWr++ = odr | RADIO_MOSI_Pin;
				*odrWr++ = odr;
			} else {
				*odrWr++ = odr;
				*odrWr++ = odr | RADIO_MOSI_Pin;
			}
		}
	}
	/*
	for (size_t i = 0; i < RADIO_TRANSMIT_TERMINATION_PULSES; i++) {
		*odrWr++ = odr | RADIO_MOSI_Pin;
		*odrWr++ = odr;
	}
	*/
	len = odrWr - odrs;

	uint32_t* odrRd = odrs;

	Radio_PowerOn();
	MAX41460_InitialProgramming();

	uint32_t csbClr = RADIO_CSB_GPIO_Port->ODR & ~(RADIO_CSB_Pin);
	uint32_t csbSet = RADIO_CSB_GPIO_Port->ODR | (RADIO_CSB_Pin);

	// stabilize CPU pipeline
	__NOP();
	__NOP();
	__NOP();

	while (len--) {
		RADIO_MOSI_GPIO_Port->ODR = *odrRd++;

		// NOP adds time padding to the instruction above.
		// Pulse duration is 50us => ~20kbps
		// With Manchester taken in account data transmission rate is ~10kbps
		__NOP(); __NOP(); __NOP(); __NOP();
	}

	RADIO_CSB_GPIO_Port->ODR = csbSet;

	// reconfigure MOSI from GPIO back to SPI
	for (int i = 0; i < 16; i++) {
		if ((1 << i) == RADIO_MOSI_Pin) {
			RADIO_MOSI_GPIO_Port->MODER = (RADIO_MOSI_GPIO_Port->MODER & (~(0x3 << (i * 2)))) | (0x2 << (i * 2));
		}
	}

	// reset MAX41460 chip
	Radio_Delay15Us();
	RADIO_CSB_GPIO_Port->ODR = csbClr;
	MAX41460_WriteByte(MAX41460_REG_CFG8);
	MAX41460_WriteByte(0x01); /* MAX41460_REG_CFG8 */
	RADIO_CSB_GPIO_Port->ODR = csbSet;

	Radio_PowerOff();

	return 0;
}
